# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of caspo.
#
# caspo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# caspo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with caspo.  If not, see <http://www.gnu.org/licenses/>.import random
# -*- coding: utf-8 -*-
import csv

def write_csv(iterable, header, filename, rowmaker=lambda r: r, verbose=True):
    f = open(filename, 'w')
    writer = csv.DictWriter(f, header)
    writer.writeheader()
    for row in iterable:
        writer.writerow(rowmaker(row))
    f.close()
    if verbose:
        print "Wrote %s" % filename

def write_models(matrix, hyperedges, outdir):
    write_csv(matrix, hyperedges, outdir + '/models.csv')

def write_frequencies(matrix, hyperedges, outdir):
    def rowmaker(row):
        nmodels = len(matrix)
        oc = 0.
    	for r in matrix:
        	oc = oc + r[row]
    	return dict(hyperedge = row, frequency=oc/nmodels)
    	
    write_csv(hyperedges, ['hyperedge','frequency'], outdir + '/frequencies.csv', rowmaker)

def write_gtt(gtt, idmodel, data, outdir):
    header = data.stimulus + map(lambda i: i+'i', data.inhibitors) + data.readouts
    write_csv(gtt, header, outdir + '/gtt-%s.csv' % idmodel, verbose=False)
    
def write_gtt_stats(models, outdir):
    rowmaker = lambda r: dict(id=r[0], mse="%.4f" % r[1], models=r[2])
    write_csv(models, ['id','mse','models'], outdir + '/gtt_stats.csv', rowmaker)
	
def write_mutuals(exclusives, inclusives, outdir):
    rowmaker = lambda r: dict(hyperedge_a = r[0], frequency_a = "%.4f" % r[1], hyperedge_b = r[2], frequency_b = "%.4f" % r[3])
    header = ['hyperedge_a', 'frequency_a', 'hyperedge_b', 'frequency_b']
    write_csv(exclusives, header, outdir + '/exclusives.csv', rowmaker)
    write_csv(inclusives, header, outdir + '/inclusives.csv', rowmaker)


# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of caspo.
#
# caspo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# caspo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with caspo.  If not, see <http://www.gnu.org/licenses/>.import random
# -*- coding: utf-8 -*-
import os
from itertools import chain, combinations

def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

def hyperedge2str(sources, target):
    neg = ''
    if sources.arg(1) == -1:
        neg = '!'
    
    hn = neg + sources.arg(0)[1:-1]
    next = sources.arg(2)
    
    while next != "end" :
        neg = ''
        if next.arg(1) == -1:
            neg = '!'
        
        hn = hn + "+" + neg + next.arg(0)[1:-1]
        next = next.arg(2)
    
    hn = hn + "=" + target
    return hn

def get_gtt(model, data):

    gtt = []
    for exp in powerset(data.stimulus + data.inhibitors):
        inputs = {}
        gtt_inputs = {}
        for s in data.stimulus:
            inputs[s] = gtt_inputs[s] = False

        for i in exp:
            if i in data.stimulus:
                inputs[i] = gtt_inputs[i] = True
            else:
                inputs[i] = gtt_inputs[i + 'i'] = False

        row = {}
        for readout in data.readouts:
            row[readout] = int(solve_node(readout, inputs, model, data))

        for k,v in gtt_inputs.iteritems():
            row[k] = int(v)
        
        gtt.append(row)

    return gtt

def solve_clause(clause, inputs, clauses, data, visited):
    value = True
    nodes = clause.arg(0)
    while nodes != "end" and value:
        sign = nodes.arg(1)
        if sign == 1:
            value = value and solve_node(nodes.arg(0)[1:-1], inputs, clauses, data, visited)
        else:
            value = value and not solve_node(nodes.arg(0)[1:-1], inputs, clauses, data, visited)
        nodes = nodes.arg(2)
    return value

def solve_node(node, inputs, clauses, data, visited=[]):
    if node in data.stimulus:
        return inputs[node]
    elif node in data.inhibitors and inputs.has_key(node):
        return inputs[node]
    elif node in visited:
        return False
    else:
        or_clauses = [c for c in clauses if c.pred() == "clause" and c.arg(2)[1:-1] == node]
        value = False
        for or_clause in or_clauses:
            value = value or solve_clause(or_clause, inputs, clauses, data, visited + [node])

        return value

def get_mse(model, data):
    rss = 0.
    obs = 0
    for e,o in zip(data.exp_conditions, data.real()):
        inputs = {}
        for inp, value in e.iteritems():
            if inp in data.stimulus:
                inputs[inp] = value == 1
            else:
                if value == 0:
                    inputs[inp] = False

        for readout, value in o.iteritems():
            val = int(solve_node(readout, inputs, model, data))

            rss = rss + pow(value - val, 2)
            obs = obs + 1

    return rss / obs
    
def get_matrix(models, hyperedges):
    matrix = []
    for model in models:
        matrixrow = {}
        m = model.to_list()
        for i in m:
            if i.pred() == "i" :
                matrixrow[hyperedge2str(i.arg(0),i.arg(2)[1:-1])] = 1
    
        ah = [h for h in hyperedges if h not in matrixrow.keys()]
        for h in ah:
            matrixrow[h] = 0
    
        matrix.append(matrixrow)

    return matrix

def clean_up():
    if os.path.isfile("parser.out"): os.remove("parser.out")
    if os.path.isfile("parsetab.py"): os.remove("parsetab.py")
    if os.path.isfile("parsetab.pyc"): os.remove("parsetab.pyc")
    if os.path.isfile("asp_py_lextab.py"): os.remove("asp_py_lextab.py")
    if os.path.isfile("asp_py_lextab.pyc"): os.remove("asp_py_lextab.pyc")
    if os.path.isfile("asp_py_parsetab.py"): os.remove("asp_py_parsetab.py")
    if os.path.isfile("asp_py_parsetab.pyc"): os.remove("asp_py_parsetab.pyc")
    if os.path.isfile("sif_parser_lextab.py"): os.remove("sif_parser_lextab.py")
    if os.path.isfile("sif_parser_lextab.pyc"): os.remove("sif_parser_lextab.pyc")



#!python
# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of caspo.
#
# caspo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# caspo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with caspo.  If not, see <http://www.gnu.org/licenses/>.import random
# -*- coding: utf-8 -*-
import os
from optparse import OptionParser
from __caspo__ import bioasp, utils, writer

if __name__ == '__main__':
    usage = "usage: %prog [options] pkn.sif midas.csv" 
    parser = OptionParser(usage)
    parser.add_option("-t", "--tolerance", dest="tolerance", type='float', default=0.,
                      help="Suboptimal enumeration tolerance: 0 <= t <= 0.5 (Default to 0)", metavar="T")

    parser.add_option("-p", "--discrete", dest="pvalue", type="int", default=2,
                      help="Discretization range exponent: 10^P (Default to 2)", metavar="P")
                      
    parser.add_option("-q", "--alpha", dest="qvalue", type="int", default=5,
                    help="Size penalty exponent 1/10^Q (Default to 5)", metavar="Q")
                    
    parser.add_option("-g", "--gtts", dest="gtts", action='store_true', default=False,
                  help="Compute Global Truth Tables (Default to False). This could take some time for many models.")

    parser.add_option("-o", "--outdir", dest="outdir", default='.',
                    help="Output directory path (Default to current directory)", metavar="O")

    (options, args) = parser.parse_args()

    if len(args) != 2:
        parser.error("incorrect number of arguments")

    allowed = [1, 2]
    if options.pvalue not in allowed:
        parser.error("incorrect value for discretization range: %s. Allowed values: %s" % (options.pvalue, allowed))

    if options.qvalue - 2*options.pvalue < 0:
        parser.error("incorrect value for discretization range %s or size penalty %s. They must satisfy q - 2p >= 0" % (options.pvalue, options.qvalue))
    
    allowed_tol = 0.5
    if options.tolerance > allowed_tol or options.tolerance < 0:
        parser.error("incorrect value for tolerance: %s. Allowed values are: 0 <= t <= %s" % (options.tolerance, allowed_tol))
        
    pkn = args[0]
    midas = args[1]
    
    models, score, optsize, data, instance = bioasp.optimize(pkn,midas,options.tolerance,options.pvalue,options.qvalue)
    
    print "\nProcessing %s boolean models..." % len(models)
    hyperedges = bioasp.get_hyperedges(instance)
    matrix = utils.get_matrix(models, hyperedges)
    
    if not os.path.exists(options.outdir):
        os.mkdir(options.outdir)
    
    writer.write_models(matrix, hyperedges, options.outdir)
    writer.write_frequencies(matrix, hyperedges, options.outdir)
    
    exclusives, inclusives = bioasp.get_mutual_hyperedges(models)
    writer.write_mutuals(exclusives, inclusives, options.outdir)
    
    if options.gtts:
        print "\nComputing models equivalences...",
        equivalences = bioasp.get_models_equivalences(models,instance)
        print 'done.'
        clusters = set(equivalences)
        print "\nComputing %s Global Truth Tables..." % len(clusters)
        muples = []
        for c in clusters:
            muples.append((c, utils.get_mse(models[c], data), equivalences.count(c)))
            writer.write_gtt(utils.get_gtt(models[c], data), c, data, options.outdir)
            
        print 'done.'
        print "Wrote %s/gtt-%s.csv" % (options.outdir, list(clusters))
        writer.write_gtt_stats(muples, options.outdir)

    utils.clean_up()



# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of caspo.
#
# caspo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# caspo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with caspo.  If not, see <http://www.gnu.org/licenses/>.import random
# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
from __caspo__ import cno, utils
from bioasp.data import psn
from bioasp.query import psnoptimization as pno
from bioasp.asp import GringoClasp, Term, TermSet

root = __file__.rsplit('/', 1)[0]

gtt_prg     =   root + '/query/gtt.lp'
mutual_prg  =   root + '/query/mutual.lp'

def get_hyperedges(instance):
	hg = pno.get_hypergraph(instance)
	hyperedges = [utils.hyperedge2str(a, a.arg(4)[1:-1]) for a in hg if a.pred() == "subset"]
	hyperedges.sort(key=len)
	
	return hyperedges


def get_mutual_hyperedges(models):
    solver = GringoClasp()
    nmodels = len(models)
    imodels = TermSet()
    imodels.add(Term('nmodels', [nmodels]))
    for i, model in enumerate(models):
        clauses = filter(lambda t: t.pred() == "clause", model)
        for c in clauses:
            imodels.add(Term('clause',[i+1] + c.arguments))

    mutuals = solver.run([ imodels.to_file(), mutual_prg ], nmodels=1, collapseTerms=False, collapseAtoms=False)
    
    exclusives = []
    inclusives = []
    for mutual in mutuals[0]:
        a = mutual.arg(0)
        b = mutual.arg(1)
        ha = utils.hyperedge2str(a.arg(0), a.arg(2)[1:-1])
        freq_a = a.arg(3) / float(nmodels)
        hb = utils.hyperedge2str(b.arg(0), b.arg(2)[1:-1])
        freq_b = b.arg(3) / float(nmodels)
        if mutual.pred() == "exclusive":
            exclusives.append((ha,freq_a,hb,freq_b))
        else:
            inclusives.append((ha,freq_a,hb,freq_b))
            
    return exclusives, inclusives
	
def get_models_equivalences(models, instance):   
    solver = GringoClasp()
    nmodels = len(models)
    eq = [-1] * nmodels
    
    for i in range(0,nmodels):
        if eq[i] == -1:
            eq[i] = i
            m1 = [c for c in models[i] if c.pred() == "clause"]
            for j in range(i+1,nmodels):
                if eq[j] == -1:
                    m2 = [c for c in models[j] if c.pred() == "clause"]
                    
                    eq_models = TermSet()
                    for c in m1:
                        eq_models.add(Term('clause',[1] + c.arguments))
    
                    for c in m2:
                        eq_models.add(Term('clause',[2] + c.arguments))

                    gtt_instance = instance.union(eq_models)
                    sat = solver.run([ gtt_instance.to_file(), gtt_prg, pno.hyper_prg ], nmodels=1, collapseTerms=False, collapseAtoms=False)
                    if len(sat) == 0:
                        eq[j] = i
    return eq
    
def optimize(pkn, midas, tolerance, pvalue, qvalue):
    print '\nReading observations',midas, '...',
    data = psn.MIDASReader(midas)
    obs = data.getTermSet(factor=pow(10,pvalue))
    print 'done.'

    print '\nReading and compressing network',pkn, '... \n\n',
    temp = "compressed_model.sif"
    cno.compress(pkn, midas, temp)
    net = psn.readSIF(temp)
    os.remove(temp)

    print '\nCreate instance...'
    instance = net.union(obs)

    print '\nSearching one global optimal model...',
    optimum = pno.get_minimal_model_size(instance, pvalue, qvalue)
    print 'done.'    

    optsize = 0
    for a in optimum[1][0] :
        if a.pred() == "clause":
            optsize = optsize + int(a.arg(1))
            
    print 'The optimal model size is %s.' % optsize

    score = int(optimum[0][0] + optimum[0][0]*tolerance)

    print 'Enumerating all models using tolerance %s and size lower than %s... ' % (tolerance,optsize),
    models = pno.get_suboptimal_models(instance, score, optsize, pvalue, qvalue)
    print 'done.'

    return models, score, optsize, data, instance



# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of caspo.
#
# caspo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# caspo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with caspo.  If not, see <http://www.gnu.org/licenses/>.import random
# -*- coding: utf-8 -*-
from setuptools import setup
                         
setup(name='caspo',
      version='1.1',
      url='http://pypi.python.org/pypi/caspo/',
      license='GPLv3',
      description='Learning Boolean Logic Models of Protein Signaling Networks powered by BioASP and CellNOptR',
      long_description=open('README.txt').read() + open('CHANGES.txt').read(),
      author='Sven Thiele, Santiago Videla',
      author_email='sthiele@irisa.fr, santiago.videla@irisa.fr',
      packages = ['__caspo__'],
      package_dir = {'__caspo__' : 'src'},
      package_data = {'__caspo__' : ['query/*.lp']},
      scripts = ['caspo.py'],
      install_requires=[
        "cellnopt.wrapper==1.0.4",
        "bioasp"
      ]
)
